import click
import configparser
from .utils import authenticate, get_config_path, get_token, get_repo_name
from tabulate import tabulate
from .classes import Github, GithubIssue


@click.group()
def cli():
    """
    A command line interface to manage all your git issues at one place
    """
    pass

@cli.command()
@click.option("-t", "--token", help="personal access token unique to you. Requires repo, read:org permissions")
def login(token):
    """
    gets token from the user to perform authenticated api calls
    """
    if token is None:
        token = click.prompt("Tip: You can generate your personal access token at https://github.com/settings/tokens\nToken", hide_input=True)
    config = configparser.ConfigParser()
    config['github-user'] = {'token': token}
    with open(get_config_path(), 'w') as configfile:
        config.write(configfile)


@cli.command()
@click.option("-r", "--repo", help = "github repository in format username/repo", default=None)
@click.option("-t", "--title", help = "issue title", default=None)
@click.option("-b", "--body", help = "description of the issue", default=None)
@click.option("-l", "--labels", help = "issue labels. add multiple labels by specifying the option again.", default=None, multiple=True)
@click.option("-a", "--assignees", help = "users to assign. add multiple usernames by specifying the option again.", default=None, multiple=True)
def create(repo, title, body, labels, assignees):
    """
    create an issue on a github repo
    """
    authenticate()
    token = get_token()
    repository = get_repo_name(repo)

    if title == None:
        title = click.prompt("An issue title is mandatory.\nTitle")

    g = Github(token)
    repo = g.get_repo(repository)
    issue = repo.create_issue(title=title, body=body, labels=labels, assignees=assignees)
    print(f"Issue #{issue.number} Created Successfully in {repository}\n\n{issue.html_url}")

    
@cli.command()
@click.argument("number", type=int)
@click.option("-r", "--repo", help = "github repository in format username/repo", default=None)
def close(repo, number):
    """
    close an issue on a github repo
    """
    authenticate()
    token = get_token()
    repository = get_repo_name(repo)

    g = Github(token)
    repo = g.get_repo(repository)
    issue = repo.get_issue(number)
    issue.close_issue()
    print(f"Issue #{issue.number} Closed Successfully in {repository}\n\n{issue.html_url}")

@cli.command()
@click.argument("number", type=int)
@click.option("-r", "--repo", help = "github repository in format username/repo", default=None)
def reopen(repo, number):
    """
    reopen an issue on a github repo
    """
    authenticate()
    token = get_token()
    repository = get_repo_name(repo)

    g = Github(token)
    repo = g.get_repo(repository)
    issue = repo.get_issue(number)
    issue.reopen_issue()
    print(f"Issue #{issue.number} Reopened Successfully in {repository}\n\n{issue.html_url}")


@cli.command()
@click.argument("number", type=int)
@click.option("-b", "--body", help = "comment body", default=None)
@click.option("-r", "--repo", help = "github repository in format username/repo", default=None)
def comment(body, repo, number):
    """
    create a new comment on a github issue
    """
    authenticate()
    token = get_token()
    repository = get_repo_name(repo)

    if body == None:
        body = click.prompt("Issue comment requires a body.\nComment")

    g = Github(token)
    repo = g.get_repo(repository)
    issue = repo.get_issue(number)
    comment = issue.create_comment(body=body)
    print(f"Comment created in issue #{issue.number} in {repository}\n\n{comment['html_url']}")


@cli.command()
@click.argument("number", type=int)
@click.option("-r", "--repo", help = "github repository in format username/repo", default=None)
@click.option("-t", "--title", help = "issue title", default=None)
@click.option("-b", "--body", help = "description of the issue", default=None)
@click.option("-s", "--state", help = "state of the issue. can be one of open or closed", type=click.Choice(['open', 'closed'], case_sensitive=False), default=None)
@click.option("-l", "--labels", help = "issue labels. add multiple labels by specifying the option again.", default=None, multiple=True)
@click.option("-a", "--assignees", help="users to assign. add multiple usernames by specifying the option again.", default=None, multiple=True)
def update(number, repo, title, body, state, labels, assignees):
    """
    update an issue on a github repo
    """
    authenticate()
    token = get_token()
    repository = get_repo_name(repo)
    
    g = Github(token)
    repo = g.get_repo(repository)
    issue = repo.get_issue(number)
    issue.update_issue(title=title, body=body, labels=labels, assignees=assignees, state=state)
    print(f"Issue #{issue.number} updated successfully in {repository}\n\n{issue.html_url}")


@cli.command()
@click.option("-r", "--repo", help = "github repository in format username/repo", default=None)
@click.option("-s", "--state", help = "state of the issue. can be one of open, closed or all",
                type=click.Choice(['open', 'closed', 'all'], case_sensitive=False), default='open')
@click.option("-a", "--author", help="filter by author", default=None)
def list(repo, state, author):
    """
    list issues of a github repo
    """
    authenticate()
    token = get_token()
    repository = get_repo_name(repo)

    g = Github(token)
    repo = g.get_repo(repository)
    table = []
    issues = repo.get_issues(params={'state':state, 'creator':author})
    for issue in issues:
        table.append(issue.get_table_attrs())
    if len(issues) == 0:
        print(f"No {'open' if state == 'all' else ''} issues found in {repository}.")
    print(tabulate(table, tablefmt="github"))

@cli.command()
@click.argument("number", type=int)
@click.option("-r", "--repo", help = "github repository in format username/repo", default=None)
@click.option("-l", "--lock-reason", help="reason to lock the issue", 
                type=click.Choice(['off topic', 'too heated', 'resolved', 'spam'], case_sensitive=False), default=None)
def lock(number, repo, lock_reason):
    """
    lock an issue. requires push access.
    must provide one of the reasons for lock
    - off topic, too heated, resolved, spam
    """
    if lock_reason is None:
        lock_reason = click.prompt("A lock reason must be specified. Lock Reason",
                                    type=click.Choice(['off topic', 'too heated', 'resolved', 'spam'], case_sensitive=False))
    authenticate()
    token = get_token()
    repository = get_repo_name(repo)

    g = Github(token)
    repo = g.get_repo(repository)
    issue = repo.get_issue(number)
    issue.lock_issue(lock_reason=lock_reason)
    print(f"Issue #{issue.number} Locked in {repository}")

@cli.command()
@click.argument("number", type=int)
@click.option("-r", "--repo", help = "github repository in format username/repo", default=None)
def unlock(number, repo):
    """
    unlock an issue.
    requires push access.
    """
    authenticate()
    token = get_token()
    repository = get_repo_name(repo)

    g = Github(token)
    repo = g.get_repo(repository)
    issue = repo.get_issue(number)
    issue.unlock_issue()
    print(f"Issue #{issue.number} unlocked in {repository}")


if __name__ == '__main__':
    cli()


"""
classes.py
contains all classes. all api calls happen here.
"""
import requests
import json
from .colour import COLOR
from datetime import datetime
import timeago
from tabulate import tabulate

class Github:
    def __init__(self, token):
        self.token = token
        self.path = None
    
    def request(self, method, path, params = None, data = None):
        url = "https://api.github.com" + path
        if params:
            url += "?"
            for key,value in params.items():
                if value != None:
                    url += f"{key}={value}"

        if data:
            data = json.dumps(data)
        headers = {"Authorization": f"token {self.token}",
                   "accept": "application/vnd.github.v3+json"}
        response = requests.request(method, url, 
                                    headers=headers,
                                    data = data)
                                    
        if response.status_code > 299:
            raise APIException(f"An API Exception has occured.\
                                \nStatus Code: {response.status_code}\nResponse: {response.text}")
        elif response.status_code == 204:
            return
        return response.json()

    def get_repo(self, name):
        self.path = f"/repos/{name}"
        data = self.request("GET", self.path)
        return GithubRepo(self, data)
        
        
class GithubRepo:
    def __init__(self, github, data):
        self.data = data
        self.github = github

    def get_issues(self, params=None):
        path = f"{self.github.path}/issues"
        data = self.github.request("GET", path, params = params)
        return [GithubIssue(self.github, d) for d in data]
    
    def get_issue(self, number):
        path = f"{self.github.path}/issues/{number}"
        data = self.github.request("GET", path)
        return GithubIssue(self.github, data)
    
    def create_issue(self, title, body=None, labels=None, assignees=None):
        data = locals()                                         # Get a dictionary of local values. At this point, all arguments of the function.
        del data['self']
        data = {k:v for k,v in data.items() if v is not None}

        path = f"{self.github.path}/issues"
        data = self.github.request("POST", path, data = data)
        return GithubIssue(self.github, data)
        
class GithubIssue:
    def __init__(self, github, data):
        self.github = github
        self.number = data['number']
        self.title = data['title']
        self.body = data['body']
        self.labels = [data['labels'][i]['name'] for i in range(len(data['labels']))] if len(data['labels']) > 0 else None
        time_obj = datetime.strptime(data['created_at'], '%Y-%m-%dT%H:%M:%SZ')
        time_diff = datetime.utcnow() - time_obj
        self.created = 'about ' + timeago.format(time_diff)
        self.assignees = [data['assignees'][i]['login'] for i in range(len(data['assignees']))] if len(data['assignees']) > 0 else None
        self.author = data['user']['login']
        self.state = data['state']
        self.html_url = data['html_url']

    def get_table_attrs(self):
        labels = ", ".join([item for item in self.labels]) if self.labels is not None else None
        return [
            f"{COLOR['GREEN']}{self.number}{COLOR['ENDC']}" if self.state == 'open' else f"{COLOR['RED']}{self.number}{COLOR['ENDC']}", 
            self.title, 
            f"{COLOR['BLUE']}{labels}{COLOR['ENDC']}" if self.labels is not None else None,
            self.created
        ]

    def close_issue(self):
        data = {'state': 'closed'}
        path = f"{self.github.path}/issues/{self.number}"
        self.github.request("PATCH", path, data = data)
    
    def reopen_issue(self):
        data = {'state': 'open'}
        path = f"{self.github.path}/issues/{self.number}"
        self.github.request("PATCH", path, data = data)

    def update_issue(self, title, body=None, labels=None, assignees=None, state=None):
        data = locals()
        del data['self']
        data = {k:v for k,v in data.items() if v is not None}

        path = f"{self.github.path}/issues/{self.number}"
        self.github.request("PATCH", path, data = data)
    
    def create_comment(self, body):
        data = {'body': body}
        path = f"{self.github.path}/issues/{self.number}/comments"
        comment = self.github.request("POST", path, data = data)
        return comment
    
    def lock_issue(self, lock_reason):
        data = {'lock_reason': lock_reason}
        path = f"{self.github.path}/issues/{self.number}/lock"
        self.github.request("PUT", path, data = data)

    def unlock_issue(self):
        path = f"{self.github.path}/issues/{self.number}/lock"
        self.github.request("DELETE", path)

class APIException(Exception):
    pass


from setuptools import setup

requirements = [
    'click',
    'certifi',
    'urllib3',
    'chardet',
    'idna',
    'requests',
    'tabulate',
    'timeago'
]

setup(
    name='gitissues',
    version='0.0.3',
    description='Manage all your git issues at one place',
    author='Roopesh V S',
    author_email='txtmeroopesh@gmail.com',
    url='https://github.com/roopeshvs/git-issues',
    packages=[
        'gitissues',
    ],
    install_requires=requirements,
    license="BSD",
    zip_safe=False,
    classifiers=[
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
    ],
    entry_points='''
        [console_scripts]
        gitissues=gitissues.cli:cli
    '''
)


"""
utils.py
contain general utility functions
"""
import os
import sys
import configparser
from subprocess import check_output, DEVNULL
from .classes import APIException
import requests
from datetime import datetime
import json
import timeago


def get_config_path():
    if sys.platform == "linux" or sys.platform == "linux2":
        return os.path.join(os.path.expanduser('~'), '.git-issues-credentials')
    elif sys.platform == "win32":
        return os.path.join(os.path.expanduser('~'), 'git-issues-credentials.ini')

def is_logged_in():
    config_file = get_config_path()
    config = configparser.ConfigParser()
    config.read(config_file)
    if 'github-user' in config.sections():
        if config['github-user']['token']:
            return True
    return False

def authenticate():
    if is_logged_in():
        pass
    else:
        print("Use the login command to login first. You will need a Personal Access Token to login.")
        exit()
        
def get_token():
    if is_logged_in():
        config_file = get_config_path()
        config = configparser.ConfigParser()
        config.read(config_file)
        return config['github-user']['token']
    
def get_repo_name(repo):
    """
    retrieves the repo for a command to work on.
    priority:
    1. --repo option
    2. Environment Variable: GITHUB_PROJECT
    3. CWD, if it is a git repo 
    """
    if repo:
        repo = repo
    if not repo:
        repo = os.environ.get("GITHUB_PROJECT", None)
    if not repo:
        if check_output(["git", "branch"]) == b"fatal: not a git repository (or any of the parent directories): .git":
            print("No git repository specified. You can specify it in three ways.\n1. with -r option\n2. GITHUB_PROJECT env variable\n3. Run the command in a git repository")
            exit()
        else:
            output = check_output(["git", "remote", "show", "origin"], stderr=DEVNULL).decode('utf=8')
            line = output.split("\n")[1]
            url_split = line.split("/")
            repo = "/".join(url_split[-2:])[:-4]
    return repo